from pydantic import ConfigDict, Field
from pydantic.dataclasses import dataclass

from sae_bench.evals.base_eval_output import (
    DEFAULT_DISPLAY,
    BaseEvalOutput,
    BaseMetricCategories,
    BaseMetrics,
    BaseResultDetail,
)
from sae_bench.evals.sparse_probing.eval_config import SparseProbingEvalConfig

EVAL_TYPE_ID_SPARSE_PROBING = "sparse_probing"


@dataclass
class SparseProbingLlmMetrics(BaseMetrics):
    llm_test_accuracy: float = Field(
        title="LLM Test Accuracy",
        description="Linear probe accuracy when training on the full LLM residual stream",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    llm_top_1_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 1 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 1 residual stream channel test accuracy",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    llm_top_2_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 2 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 2 residual stream channels test accuracy",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    llm_top_5_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 5 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 5 residual stream channels test accuracy",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    llm_top_10_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 10 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 10 residual stream channels",
    )
    llm_top_20_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 20 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 20 residual stream channels",
    )
    llm_top_50_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 50 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 50 residual stream channels",
    )
    llm_top_100_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 100 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 100 residual stream channels",
    )


@dataclass
class SparseProbingSaeMetrics(BaseMetrics):
    sae_test_accuracy: float | None = Field(
        default=None,
        title="SAE Test Accuracy",
        description="Linear probe accuracy when trained on all SAE latents",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    sae_top_1_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 1 Test Accuracy",
        description="Linear probe accuracy when trained on the top 1 SAE latents",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    sae_top_2_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 2 Test Accuracy",
        description="Linear probe accuracy when trained on the top 2 SAE latents",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    sae_top_5_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 5 Test Accuracy",
        description="Linear probe accuracy when trained on the top 5 SAE latents",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    sae_top_10_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 10 Test Accuracy",
        description="Linear probe accuracy when trained on the top 10 SAE latents",
    )
    sae_top_20_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 20 Test Accuracy",
        description="Linear probe accuracy when trained on the top 20 SAE latents",
    )
    sae_top_50_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 50 Test Accuracy",
        description="Linear probe accuracy when trained on the top 50 SAE latents",
    )
    sae_top_100_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 100 Test Accuracy",
        description="Linear probe accuracy when trained on the top 100 SAE latents",
    )


@dataclass
class SparseProbingMetricCategories(BaseMetricCategories):
    llm: SparseProbingLlmMetrics = Field(
        title="LLM",
        description="LLM metrics",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    sae: SparseProbingSaeMetrics = Field(
        title="SAE",
        description="SAE metrics",
        json_schema_extra=DEFAULT_DISPLAY,
    )


@dataclass
class SparseProbingResultDetail(BaseResultDetail):
    dataset_name: str = Field(
        title="Dataset Name",
        description="Dataset name",
    )

    llm_test_accuracy: float = Field(
        title="LLM Test Accuracy",
        description="Linear probe accuracy when trained on all LLM residual stream channels",
    )
    llm_top_1_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 1 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 1 residual stream channels",
    )
    llm_top_2_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 2 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 2 residual stream channels",
    )
    llm_top_5_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 5 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 5 residual stream channels",
    )
    llm_top_10_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 10 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 10 residual stream channels",
    )
    llm_top_20_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 20 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 20 residual stream channels",
    )
    llm_top_50_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 50 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 50 residual stream channels",
    )
    llm_top_100_test_accuracy: float | None = Field(
        default=None,
        title="LLM Top 100 Test Accuracy",
        description="Linear probe accuracy when trained on the LLM top 100 residual stream channels",
    )

    sae_test_accuracy: float | None = Field(
        default=None,
        title="SAE Test Accuracy",
        description="Linear probe accuracy when trained on all SAE latents",
    )
    sae_top_1_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 1 Test Accuracy",
        description="Linear probe accuracy when trained on the top 1 SAE latents",
    )
    sae_top_2_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 2 Test Accuracy",
        description="Linear probe accuracy when trained on the top 2 SAE latents",
    )
    sae_top_5_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 5 Test Accuracy",
        description="Linear probe accuracy when trained on the top 5 SAE latents",
    )
    sae_top_10_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 10 Test Accuracy",
        description="Linear probe accuracy when trained on the top 10 SAE latents",
    )
    sae_top_20_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 20 Test Accuracy",
        description="Linear probe accuracy when trained on the top 20 SAE latents",
    )
    sae_top_50_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 50 Test Accuracy",
        description="Linear probe accuracy when trained on the top 50 SAE latents",
    )
    sae_top_100_test_accuracy: float | None = Field(
        default=None,
        title="SAE Top 100 Test Accuracy",
        description="Linear probe accuracy when trained on the top 100 SAE latents",
    )


@dataclass(config=ConfigDict(title="Sparse Probing"))
class SparseProbingEvalOutput(
    BaseEvalOutput[
        SparseProbingEvalConfig,
        SparseProbingMetricCategories,
        SparseProbingResultDetail,
    ]
):
    # This will end up being the description of the eval in the UI.
    """
    An evaluation using SAEs to probe for supervised concepts in LLMs. We use sparse probing with the top K SAE latents and probe for over 30 different classes across 5 datasets.
    """

    eval_config: SparseProbingEvalConfig
    eval_id: str
    datetime_epoch_millis: int
    eval_result_metrics: SparseProbingMetricCategories
    eval_result_details: list[SparseProbingResultDetail] = Field(
        default_factory=list,
        title="Per-Dataset Sparse Probing Results",
        description="Each object is a stat on the sparse probing results for a dataset.",
    )
    eval_type_id: str = Field(default=EVAL_TYPE_ID_SPARSE_PROBING)



# %%

# ruff: noqa
# type: ignore
# TODO: Add proper type hints and enable linting, most imports are broken currently


import copy
import random
import time
from dataclasses import asdict

import pandas as pd
import torch
from sae_lens import SAE
from sae_lens.toolkit.pretrained_saes_directory import get_pretrained_saes_directory
from tqdm import tqdm
from transformer_lens import HookedTransformer

import sae_bench.sae_bench_utils.activation_collection as activation_collection
import sae_bench.sae_bench_utils.dataset_utils as dataset_utils
import sae_bench.sae_bench_utils.formatting_utils as formatting_utils
from sae_bench.evals.sparse_probing import eval_config, probe_training


def average_test_accuracy(test_accuracies: dict[str, float]) -> float:
    return sum(test_accuracies.values()) / len(test_accuracies)


start_time = time.time()

if torch.backends.mps.is_available():
    device = "mps"
else:
    device = "cuda" if torch.cuda.is_available() else "cpu"

print(f"Using device: {device}")

config = eval_config.EvalConfig()

random.seed(config.random_seed)
torch.manual_seed(config.random_seed)

# populate selected_saes_dict
for release in config.sae_releases:
    if "gemma-scope" in release:
        config.selected_saes_dict[release] = (
            formatting_utils.find_gemmascope_average_l0_sae_names(config.layer)
        )
    else:
        config.selected_saes_dict[release] = formatting_utils.filter_sae_names(
            sae_names=release,
            layers=[config.layer],
            include_checkpoints=config.include_checkpoints,
            trainer_ids=config.trainer_ids,
        )

    print(f"SAE release: {release}, SAEs: {config.selected_saes_dict[release]}")

# %%

# TODO: Make this nicer.
sae_map_df = pd.DataFrame.from_records(
    {k: v.__dict__ for k, v in get_pretrained_saes_directory().items()}
).T

results_dict = {}
results_dict["custom_eval_results"] = {}

llm_batch_size = activation_collection.LLM_NAME_TO_BATCH_SIZE[config.model_name]
llm_dtype = activation_collection.LLM_NAME_TO_DTYPE[config.model_name]

model = HookedTransformer.from_pretrained_no_processing(
    config.model_name, device=device, dtype=llm_dtype
)

train_df, test_df = dataset_utils.load_huggingface_dataset(config.dataset_name)
train_data, test_data = dataset_utils.get_multi_label_train_test_data(
    train_df,
    test_df,
    config.dataset_name,
    config.probe_train_set_size,
    config.probe_test_set_size,
    config.random_seed,
)
dataset_utils
train_data = dataset_utils.filter_dataset(train_data, config.chosen_classes)
test_data = dataset_utils.filter_dataset(test_data, config.chosen_classes)

train_data = dataset_utils.tokenize_data(
    train_data, model.tokenizer, config.context_length, device
)
test_data = dataset_utils.tokenize_data(
    test_data, model.tokenizer, config.context_length, device
)

print(f"Running evaluation for layer {config.layer}")
hook_name = f"blocks.{config.layer}.hook_resid_post"

all_train_acts_BLD = activation_collection.get_all_llm_activations(
    train_data, model, llm_batch_size, hook_name
)
all_test_acts_BLD = activation_collection.get_all_llm_activations(
    test_data, model, llm_batch_size, hook_name
)

all_train_acts_BD = activation_collection.create_meaned_model_activations(
    all_train_acts_BLD
)
all_test_acts_BD = activation_collection.create_meaned_model_activations(
    all_test_acts_BLD
)

llm_probes, llm_test_accuracies = probe_training.train_probe_on_activations(
    all_train_acts_BD,
    all_test_acts_BD,
    select_top_k=None,
)

llm_results = {"llm_test_accuracy": average_test_accuracy(llm_test_accuracies)}

# %%


import torch
import torch.nn as nn
from beartype import beartype
from jaxtyping import Bool, Float, Int, jaxtyped
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score

import sae_bench.sae_bench_utils.dataset_info as dataset_info


class Probe(nn.Module):
    def __init__(self, activation_dim: int, dtype: torch.dtype):
        super().__init__()
        self.net = nn.Linear(activation_dim, 1, bias=True, dtype=dtype)

    def forward(self, x):
        return self.net(x).squeeze(-1)


@jaxtyped(typechecker=beartype)
def prepare_probe_data(
    all_activations: dict[str, Float[torch.Tensor, "num_datapoints d_model"]],
    class_idx: str,
) -> tuple[Float[torch.Tensor, "batch_size d_model"], Int[torch.Tensor, "batch_size"]]:
    positive_acts_BD = all_activations[class_idx]
    device = positive_acts_BD.device

    num_positive = len(positive_acts_BD)

    # Collect all negative class activations and labels
    negative_acts = []
    for idx, acts in all_activations.items():
        if idx != class_idx:
            negative_acts.append(acts)

    negative_acts = torch.cat(negative_acts)

    # Randomly select num_positive samples from negative class
    indices = torch.randperm(len(negative_acts))[:num_positive]
    selected_negative_acts_BD = negative_acts[indices]

    assert selected_negative_acts_BD.shape == positive_acts_BD.shape

    # Combine positive and negative samples
    combined_acts = torch.cat([positive_acts_BD, selected_negative_acts_BD])

    combined_labels = torch.empty(len(combined_acts), dtype=torch.int, device=device)
    combined_labels[:num_positive] = dataset_info.POSITIVE_CLASS_LABEL
    combined_labels[num_positive:] = dataset_info.NEGATIVE_CLASS_LABEL

    # Shuffle the combined data
    shuffle_indices = torch.randperm(len(combined_acts))
    shuffled_acts = combined_acts[shuffle_indices]
    shuffled_labels = combined_labels[shuffle_indices]

    return shuffled_acts, shuffled_labels


@jaxtyped(typechecker=beartype)
def get_top_k_mean_diff_mask(
    acts_BD: Float[torch.Tensor, "batch_size d_model"],
    labels_B: Int[torch.Tensor, "batch_size"],
    k: int,
) -> Bool[torch.Tensor, "k"]:
    positive_mask_B = labels_B == dataset_info.POSITIVE_CLASS_LABEL
    negative_mask_B = labels_B == dataset_info.NEGATIVE_CLASS_LABEL

    positive_distribution_D = acts_BD[positive_mask_B].mean(dim=0)
    negative_distribution_D = acts_BD[negative_mask_B].mean(dim=0)
    distribution_diff_D = (positive_distribution_D - negative_distribution_D).abs()
    top_k_indices_D = torch.argsort(distribution_diff_D, descending=True)[:k]

    mask_D = torch.ones(acts_BD.shape[1], dtype=torch.bool, device=acts_BD.device)
    mask_D[top_k_indices_D] = False

    return mask_D


@jaxtyped(typechecker=beartype)
def apply_topk_mask_gpu(
    acts_BD: Float[torch.Tensor, "batch_size d_model"],
    mask_D: Bool[torch.Tensor, "d_model"],
) -> Float[torch.Tensor, "batch_size k"]:
    masked_acts_BD = acts_BD.clone()
    masked_acts_BD[:, mask_D] = 0.0

    return masked_acts_BD


@jaxtyped(typechecker=beartype)
def apply_topk_mask_sklearn(
    acts_BD: Float[torch.Tensor, "batch_size d_model"],
    mask_D: Bool[torch.Tensor, "d_model"],
) -> Float[torch.Tensor, "batch_size k"]:
    masked_acts_BD = acts_BD.clone()

    masked_acts_BD = masked_acts_BD[:, ~mask_D]

    return masked_acts_BD


@beartype
def train_sklearn_probe(
    train_inputs: Float[torch.Tensor, "train_dataset_size d_model"],
    train_labels: Int[torch.Tensor, "train_dataset_size"],
    test_inputs: Float[torch.Tensor, "test_dataset_size d_model"],
    test_labels: Int[torch.Tensor, "test_dataset_size"],
    max_iter: int = 100,  # non-default sklearn value, increased due to convergence warnings
    C: float = 1.0,  # default sklearn value
    verbose: bool = False,
    l1_ratio: float | None = None,
) -> tuple[LogisticRegression, float]:
    # Convert torch tensors to numpy arrays
    train_inputs_np = train_inputs.cpu().numpy()
    train_labels_np = train_labels.cpu().numpy()
    test_inputs_np = test_inputs.cpu().numpy()
    test_labels_np = test_labels.cpu().numpy()

    # Initialize the LogisticRegression model
    if l1_ratio is not None:
        # Use Elastic Net regularization
        probe = LogisticRegression(
            penalty="elasticnet",
            solver="saga",
            C=C,
            l1_ratio=l1_ratio,
            max_iter=max_iter,
            verbose=int(verbose),
        )
    else:
        # Use L2 regularization
        probe = LogisticRegression(
            penalty="l2", C=C, max_iter=max_iter, verbose=int(verbose)
        )

    # Train the model
    probe.fit(train_inputs_np, train_labels_np)

    # Compute accuracies
    train_accuracy = accuracy_score(train_labels_np, probe.predict(train_inputs_np))
    test_accuracy = accuracy_score(test_labels_np, probe.predict(test_inputs_np))

    if verbose:
        print("\nTraining completed.")
        print(f"Train accuracy: {train_accuracy}, Test accuracy: {test_accuracy}\n")

    return probe, test_accuracy


# Helper function to test the probe
@beartype
def test_sklearn_probe(
    inputs: Float[torch.Tensor, "dataset_size d_model"],
    labels: Int[torch.Tensor, "dataset_size"],
    probe: LogisticRegression,
) -> float:
    inputs_np = inputs.cpu().numpy()
    labels_np = labels.cpu().numpy()
    predictions = probe.predict(inputs_np)
    return accuracy_score(labels_np, predictions)


@jaxtyped(typechecker=beartype)
@torch.no_grad
def test_probe_gpu(
    inputs: Float[torch.Tensor, "test_dataset_size d_model"],
    labels: Int[torch.Tensor, "test_dataset_size"],
    batch_size: int,
    probe: Probe,
) -> float:
    criterion = nn.BCEWithLogitsLoss()

    with torch.no_grad():
        corrects_0 = []
        corrects_1 = []
        all_corrects = []
        losses = []

        for i in range(0, len(labels), batch_size):
            acts_BD = inputs[i : i + batch_size]
            labels_B = labels[i : i + batch_size]
            logits_B = probe(acts_BD)
            preds_B = (logits_B > 0.0).long()
            correct_B = (preds_B == labels_B).float()

            all_corrects.append(correct_B)
            corrects_0.append(correct_B[labels_B == 0])
            corrects_1.append(correct_B[labels_B == 1])

            loss = criterion(logits_B, labels_B.to(dtype=probe.net.weight.dtype))
            losses.append(loss)

        accuracy_all = torch.cat(all_corrects).mean().item()
        accuracy_0 = torch.cat(corrects_0).mean().item() if corrects_0 else 0.0
        accuracy_1 = torch.cat(corrects_1).mean().item() if corrects_1 else 0.0
        all_loss = torch.stack(losses).mean().item()

    return accuracy_all


@jaxtyped(typechecker=beartype)
def train_probe_gpu(
    train_inputs: Float[torch.Tensor, "train_dataset_size d_model"],
    train_labels: Int[torch.Tensor, "train_dataset_size"],
    test_inputs: Float[torch.Tensor, "test_dataset_size d_model"],
    test_labels: Int[torch.Tensor, "test_dataset_size"],
    dim: int,
    batch_size: int,
    epochs: int,
    lr: float,
    verbose: bool = False,
    l1_penalty: float | None = None,
    early_stopping_patience: int = 5,
):  # tuple[Probe, float]:
    device = train_inputs.device
    model_dtype = train_inputs.dtype

    print(f"Training probe with dim: {dim}, device: {device}, dtype: {model_dtype}")

    probe = Probe(dim, model_dtype).to(device)
    optimizer = torch.optim.AdamW(probe.parameters(), lr=lr)
    criterion = nn.BCEWithLogitsLoss()

    best_test_accuracy = -1.0
    best_probe = None
    patience_counter = 0

    for epoch in range(epochs):
        for i in range(0, len(train_inputs), batch_size):
            acts_BD = train_inputs[i : i + batch_size]
            labels_B = train_labels[i : i + batch_size]
            logits_B = probe(acts_BD)
            loss = criterion(
                logits_B, labels_B.clone().detach().to(device=device, dtype=model_dtype)
            )

            if l1_penalty is not None:
                l1_loss = l1_penalty * torch.sum(torch.abs(probe.net.weight))
                loss += l1_loss

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

        train_accuracy = test_probe_gpu(train_inputs, train_labels, batch_size, probe)
        test_accuracy = test_probe_gpu(test_inputs, test_labels, batch_size, probe)

        if test_accuracy > best_test_accuracy:
            best_test_accuracy = test_accuracy
            best_probe = copy.deepcopy(probe)
            patience_counter = 0
        else:
            patience_counter += 1

        if verbose:
            print(
                f"Epoch {epoch + 1}/{epochs} Loss: {loss.item()}, train accuracy: {train_accuracy}, test accuracy: {test_accuracy}"
            )

        if patience_counter >= early_stopping_patience:
            print(f"Early stopping triggered after {epoch + 1} epochs")
            break

    print(type(best_probe))

    return best_probe, best_test_accuracy


@jaxtyped(typechecker=beartype)
def train_probe_on_activations(
    train_activations: dict[str, Float[torch.Tensor, "train_dataset_size d_model"]],
    test_activations: dict[str, Float[torch.Tensor, "test_dataset_size d_model"]],
    select_top_k: int | None = None,
):  # -> tuple[dict[str, LogisticRegression], dict[str, float]]:
    torch.set_grad_enabled(True)

    probes, test_accuracies = {}, {}

    for profession in train_activations.keys():
        train_acts, train_labels = prepare_probe_data(train_activations, profession)

        test_acts, test_labels = prepare_probe_data(test_activations, profession)

        if select_top_k is not None:
            activation_mask_D = get_top_k_mean_diff_mask(
                train_acts, train_labels, select_top_k
            )
            train_acts = apply_topk_mask_sklearn(train_acts, activation_mask_D)
            test_acts = apply_topk_mask_sklearn(test_acts, activation_mask_D)

            # train_acts = apply_topk_mask_gpu(train_acts, activation_mask_D)
            # test_acts = apply_topk_mask_gpu(test_acts, activation_mask_D)

        activation_dim = train_acts.shape[1]

        print(f"Num non-zero elements: {activation_dim}")

        probe, test_accuracy = train_sklearn_probe(
            train_acts,
            train_labels,
            test_acts,
            test_labels,
            verbose=False,
        )

        # probe, test_accuracy = train_probe_gpu(
        #     train_acts,
        #     train_labels,
        #     test_acts,
        #     test_labels,
        #     dim=activation_dim,
        #     batch_size=1000,
        #     epochs=100,
        #     lr=1e-2,
        #     verbose=False,
        #     early_stopping_patience=10,
        # )

        print(f"\nTest accuracy for {profession}: {test_accuracy}")

        probes[profession] = probe
        test_accuracies[profession] = test_accuracy

    return probes, test_accuracies


# %%

random.seed(config.random_seed)
torch.manual_seed(config.random_seed)

config.k_values = [100]

for k in config.k_values:
    llm_top_k_probes, llm_top_k_test_accuracies = train_probe_on_activations(
        all_train_acts_BD,
        all_test_acts_BD,
        select_top_k=k,
    )
    llm_results[f"llm_top_{k}_test_accuracy"] = average_test_accuracy(
        llm_top_k_test_accuracies
    )
    print(f"Top {k} test accuracy: {llm_results[f'llm_top_{k}_test_accuracy']}")

# %%

sae_release = None

for sae_release in config.selected_saes_dict:
    sae_release = sae_release
print(
    f"Running evaluation for SAE release: {sae_release}, SAEs: {config.selected_saes_dict[sae_release]}"
)
sae_id_to_name_map = sae_map_df.saes_map[sae_release]
sae_name_to_id_map = {v: k for k, v in sae_id_to_name_map.items()}

for sae_name in tqdm(
    config.selected_saes_dict[sae_release],
    desc="Running SAE evaluation on all selected SAEs",
):
    sae_id = sae_name_to_id_map[sae_name]

    sae, cfg_dict, sparsity = SAE.from_pretrained(
        release=sae_release,
        sae_id=sae_id,
        device=device,
    )
    sae = sae.to(device=device)

# %%

start_time = time.time()
config.sae_batch_size = 125

all_sae_train_acts_BF = activation_collection.get_sae_meaned_activations(
    all_train_acts_BLD, sae, config.sae_batch_size, llm_dtype
)
all_sae_test_acts_BF = activation_collection.get_sae_meaned_activations(
    all_test_acts_BLD, sae, config.sae_batch_size, llm_dtype
)

end_time = time.time()
print(f"Time taken: {end_time - start_time}")

# %%

start_time = time.time()

sae_probes, sae_test_accuracies = train_probe_on_activations(
    all_sae_train_acts_BF,
    all_sae_test_acts_BF,
    select_top_k=None,
)

end_time = time.time()
print(f"Time taken: {end_time - start_time}")

start_time = time.time()

results_dict["custom_eval_results"][sae_name] = {}

for llm_result_key, llm_result_value in llm_results.items():
    results_dict["custom_eval_results"][sae_name][llm_result_key] = llm_result_value

# results_dict["custom_eval_results"][sae_name]["sae_test_accuracy"] = average_test_accuracy(
#     sae_test_accuracies
# )

config.k_values = [1, 2, 5]

for k in config.k_values:
    sae_top_k_probes, sae_top_k_test_accuracies = train_probe_on_activations(
        all_sae_train_acts_BF,
        all_sae_test_acts_BF,
        select_top_k=k,
    )
    results_dict["custom_eval_results"][sae_name][f"sae_top_{k}_test_accuracy"] = (
        average_test_accuracy(sae_top_k_test_accuracies)
    )

results_dict["custom_eval_config"] = asdict(config)

end_time = time.time()
print(f"Time taken: {end_time - start_time}")


# %%
print(sae_top_k_probes["0"])
# %%



"""
Shared helpers for experiments
"""

import re
from collections.abc import Iterable
from pathlib import Path
from typing import Callable, Literal

import numpy as np
import pandas as pd
import torch
from tqdm.autonotebook import tqdm
from transformer_lens import HookedTransformer
from transformers import PreTrainedTokenizerFast

from sae_bench.evals.absorption.probing import (
    LinearProbe,
    create_dataset_probe_training,
    gen_and_save_df_acts_probing,
    save_probe_and_data,
    train_linear_probe_for_task,
)
from sae_bench.evals.absorption.prompting import (
    Formatter,
    first_letter_formatter,
)
from sae_bench.evals.absorption.vocab import get_alpha_tokens

DEFAULT_DTYPE = torch.bfloat16 if torch.cuda.is_available() else torch.float32
DEFAULT_DEVICE = "cuda" if torch.cuda.is_available() else "cpu"

RESULTS_DIR = Path(__file__).parent.parent.parent / "artifacts" / "absorption"
PROBES_DIR = RESULTS_DIR / "probes"


def dtype_to_str(dtype: torch.dtype | str) -> str:
    return str(dtype).replace("torch.", "")


def load_or_train_probe(
    model: HookedTransformer,
    base_template: str,
    pos_idx: int,
    layer: int = 0,
    probes_dir: str | Path = PROBES_DIR,
    dtype: torch.dtype = DEFAULT_DTYPE,
    device: str = DEFAULT_DEVICE,
) -> LinearProbe:
    probe_path = (
        Path(probes_dir) / f"{model.cfg.model_name}" / f"layer_{layer}" / "probe.pth"
    )
    if not probe_path.exists():
        print(f"Probe for layer {layer} not found, training...")
        train_and_save_probes(
            model,
            layers=[layer],
            probes_dir=probes_dir,
            base_template=base_template,
            pos_idx=pos_idx,
            device=device,
        )
    return load_probe(model.cfg.model_name, layer, probes_dir, dtype, device)


def load_probe(
    model_name: str | None,
    layer: int = 0,
    probes_dir: str | Path = PROBES_DIR,
    dtype: torch.dtype = DEFAULT_DTYPE,
    device: str = DEFAULT_DEVICE,
) -> LinearProbe:
    probe = torch.load(
        Path(probes_dir) / f"{model_name}" / f"layer_{layer}" / "probe.pth",
        map_location=device,
        weights_only=False,
    ).to(dtype=dtype)
    return probe


def load_probe_data_split_or_train(
    model: HookedTransformer,
    base_template: str,
    pos_idx: int,
    layer: int = 0,
    split: Literal["train", "test"] = "test",
    probes_dir: str | Path = PROBES_DIR,
    dtype: torch.dtype = DEFAULT_DTYPE,
    device: str = DEFAULT_DEVICE,
) -> tuple[torch.Tensor, list[tuple[str, int]]]:
    probe_path = (
        Path(probes_dir) / f"{model.cfg.model_name}" / f"layer_{layer}" / "probe.pth"
    )
    if not probe_path.exists():
        print(f"Probe for layer {layer} not found, training...")
        train_and_save_probes(
            model,
            layers=[layer],
            probes_dir=probes_dir,
            base_template=base_template,
            pos_idx=pos_idx,
            device=device,
        )
    return load_probe_data_split(
        model,  # type: ignore
        layer,
        split,
        probes_dir,
        dtype,
        device,
    )


@torch.inference_mode()
def load_probe_data_split(
    model: HookedTransformer,
    layer: int = 0,
    split: Literal["train", "test"] = "test",
    probes_dir: str | Path = PROBES_DIR,
    dtype: torch.dtype = DEFAULT_DTYPE,
    device: str = DEFAULT_DEVICE,
) -> tuple[torch.Tensor, list[tuple[str, int]]]:
    np_data = np.load(
        Path(probes_dir) / f"{model.cfg.model_name}" / f"layer_{layer}" / "data.npz",
    )
    df = pd.read_csv(
        Path(probes_dir)
        / f"{model.cfg.model_name}"
        / f"layer_{layer}"
        / f"{split}_df.csv",
        keep_default_na=False,
        na_values=[""],
    )
    activations = torch.from_numpy(np_data[f"X_{split}"]).to(device, dtype=dtype)
    labels = np_data[f"y_{split}"].tolist()
    return _parse_probe_data_split(
        model.tokenizer,  # type: ignore
        activations,
        split_labels=labels,
        df=df,  # type: ignore
    )


def _parse_probe_data_split(
    tokenizer: PreTrainedTokenizerFast,
    split_activations: torch.Tensor,
    split_labels: list[int],
    df: pd.DataFrame,
) -> tuple[torch.Tensor, list[tuple[str, int]]]:
    valid_act_indices = []
    vocab_with_labels = []
    raw_tokens_with_labels = [
        (df.iloc[idx]["token"], label) for idx, label in enumerate(split_labels)
    ]
    for idx, (token, label) in enumerate(raw_tokens_with_labels):
        # sometimes we have tokens that look like <0x6A>
        if not isinstance(token, str) or re.match(r"[\d<>]", token):
            continue
        vocab_with_labels.append((tokenizer.convert_tokens_to_string([token]), label))
        valid_act_indices.append(idx)
    activations = split_activations[valid_act_indices]
    if activations.shape[0] != len(vocab_with_labels):
        raise ValueError(
            f"Activations and vocab with labels have different lengths: "
            f"{activations.shape[0]} != {len(vocab_with_labels)}"
        )
    return activations.clone(), vocab_with_labels


def get_or_make_dir(
    experiment_dir: str | Path,
) -> Path:
    """
    Helper to create a directory for a specific task within an experiment directory.
    """
    experiment_dir = Path(experiment_dir)
    experiment_dir.mkdir(parents=True, exist_ok=True)
    return experiment_dir


def load_experiment_df(
    experiment_name: str,
    path: Path,
) -> pd.DataFrame:
    """
    Helper to load a DF or error if it doesn't exist.
    """
    if not path.exists():
        raise FileNotFoundError(
            f"{path} does not exist. Run the {experiment_name} experiment first."
        )
    return pd.read_parquet(path)


def load_df_or_run(
    fn: Callable[[], pd.DataFrame],
    path: Path,
    force: bool = False,
) -> pd.DataFrame:
    return load_dfs_or_run(lambda: [fn()], [path], force)[0]


def load_dfs_or_run(
    fn: Callable[[], Iterable[pd.DataFrame]],
    paths: Iterable[Path],
    force: bool = False,
) -> list[pd.DataFrame]:
    if force or not all(path.exists() for path in paths):
        dfs = fn()
        for df, path in zip(dfs, paths):
            path.parent.mkdir(parents=True, exist_ok=True)
            df.to_parquet(path, index=False)
    else:
        print(f"{paths} exist(s), loading from disk")
        dfs = [pd.read_parquet(path) for path in paths]
    return list(dfs)


def create_and_train_probe(
    model: HookedTransformer,
    formatter: Formatter,
    hook_point: str,
    probes_dir: str | Path,
    vocab: list[str],
    batch_size: int,
    num_epochs: int,
    lr: float,
    device: torch.device,
    base_template: str,
    pos_idx: int,
    num_prompts_per_token: int = 1,
):
    train_dataset, test_dataset = create_dataset_probe_training(
        vocab=vocab,
        formatter=formatter,
        num_prompts_per_token=num_prompts_per_token,
        base_template=base_template,
    )

    layer = int(hook_point.split(".")[1])

    train_df, test_df, train_activations, test_activations = (
        gen_and_save_df_acts_probing(
            model=model,
            train_dataset=train_dataset,
            test_dataset=test_dataset,
            path=Path(probes_dir) / f"{model.cfg.model_name}" / f"layer_{layer}",
            hook_point=hook_point,
            batch_size=batch_size,
            position_idx=pos_idx,
        )
    )

    num_classes = 26
    probe, probe_data = train_linear_probe_for_task(
        train_df=train_df,
        test_df=test_df,
        train_activations=train_activations,
        test_activations=test_activations,
        num_classes=num_classes,
        batch_size=batch_size,
        num_epochs=num_epochs,
        lr=lr,
        device=device,
    )
    save_probe_and_data(
        probe,
        probe_data,
        probing_path=Path(probes_dir) / f"{model.cfg.model_name}" / f"layer_{layer}",
    )


def train_and_save_probes(
    model: HookedTransformer,
    layers: list[int],
    base_template: str,
    pos_idx: int,
    probes_dir: str | Path = PROBES_DIR,
    batch_size=64,
    num_epochs=50,
    lr=1e-2,
    device=DEFAULT_DEVICE,
):
    vocab = get_alpha_tokens(model.tokenizer)  # type: ignore
    for layer in tqdm(layers):
        hook_point = f"blocks.{layer}.hook_resid_post"
        create_and_train_probe(
            model=model,
            hook_point=hook_point,
            formatter=first_letter_formatter(),
            probes_dir=probes_dir,
            vocab=vocab,
            batch_size=batch_size,
            num_epochs=num_epochs,
            lr=lr,
            device=torch.device(device),
            base_template=base_template,
            pos_idx=pos_idx,
        )



import os
import random
from dataclasses import dataclass
from math import exp, log
from pathlib import Path
from typing import Callable, Literal

import numpy as np
import pandas as pd
import torch
from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score
from torch import nn, optim
from torch.nn.functional import one_hot
from torch.utils.data import DataLoader, TensorDataset

# Autocheck if the instance is a notebook or not (fixes weird bugs in colab)
from tqdm.autonotebook import tqdm
from transformer_lens import HookedTransformer

from sae_bench.evals.absorption.prompting import (
    Formatter,
    SpellingPrompt,
    create_icl_prompt,
)
from sae_bench.evals.absorption.util import DEFAULT_DEVICE, batchify
from sae_bench.evals.absorption.vocab import LETTERS


class LinearProbe(nn.Module):
    """
    Based on by https://github.com/jbloomAus/alphabetical_probe/blob/main/src/probes.py
    """

    def __init__(self, input_dim, num_outputs: int = 1):
        super().__init__()
        self.fc = nn.Linear(input_dim, num_outputs)

    def forward(self, x):
        return self.fc(x)

    @property
    def weights(self):
        return self.fc.weight

    @property
    def biases(self):
        return self.fc.bias


def _calc_pos_weights(y: torch.Tensor) -> torch.Tensor:
    num_pos_samples = y.sum(dim=0)
    num_neg_samples = len(y) - num_pos_samples
    return num_neg_samples / num_pos_samples


def train_multi_probe(
    x_train: torch.Tensor,  # tensor of shape (num_samples, input_dim)
    y_train: torch.Tensor,  # tensor of shape (num_samples, num_probes), with values in [0, 1]
    num_probes: int | None = None,  # inferred from y_train if None
    batch_size: int = 4096,
    num_epochs: int = 100,
    lr: float = 0.01,
    end_lr: float = 1e-5,
    weight_decay: float = 1e-6,
    show_progress: bool = True,
    optimizer: Literal["Adam", "SGD", "AdamW"] = "Adam",
    extra_loss_fn: (
        Callable[[LinearProbe, torch.Tensor, torch.Tensor], torch.Tensor] | None
    ) = None,
    verbose: bool = False,
    device: torch.device = DEFAULT_DEVICE,
    map_acts: Callable[[torch.Tensor], torch.Tensor] | None = None,
    probe_dim: int | None = None,
) -> LinearProbe:
    """
    Train a multi-class one-vs-rest logistic regression probe on the given data.
    This is equivalent to training num_probes separate binary logistic regression probes.

    Args:
        x_train: tensor of shape (num_samples, input_dim)
        y_train: one_hot (or multi-hot) tensor of shape (num_samples, num_probes), with values in [0, 1]
        num_probes: number of probes to train simultaneously
        batch_size: batch size for training
        num_epochs: number of epochs to train for
        lr: learning rate
        weight_decay: weight decay
        show_progress: whether to show a progress bar
        device: device to train on
    """
    dtype = x_train.dtype
    num_probes = num_probes or y_train.shape[-1]
    dataset = TensorDataset(x_train, y_train.to(dtype=dtype))
    loader = DataLoader(dataset, batch_size=batch_size, shuffle=True)
    if probe_dim is None:
        probe_dim = x_train.shape[-1]
    probe = LinearProbe(probe_dim, num_outputs=num_probes).to(device, dtype=dtype)

    _run_probe_training(
        probe,
        loader,
        loss_fn=nn.BCEWithLogitsLoss(pos_weight=_calc_pos_weights(y_train).to(device)),
        num_epochs=num_epochs,
        lr=lr,
        end_lr=end_lr,
        weight_decay=weight_decay,
        show_progress=show_progress,
        optimizer_name=optimizer,
        extra_loss_fn=extra_loss_fn,
        verbose=verbose,
        device=device,
        map_acts=map_acts,
    )

    return probe


def train_binary_probe(
    x_train: torch.Tensor,  # tensor of shape (num_samples, input_dim)
    y_train: torch.Tensor,  # tensor of shape (num_samples,), with values in [0, 1]
    batch_size: int = 256,
    num_epochs: int = 100,
    lr: float = 0.01,
    end_lr: float = 1e-5,
    weight_decay: float = 1e-6,
    show_progress: bool = True,
    optimizer: Literal["Adam", "SGD", "AdamW"] = "Adam",
    extra_loss_fn: (
        Callable[[LinearProbe, torch.Tensor, torch.Tensor], torch.Tensor] | None
    ) = None,
    verbose: bool = False,
    device: torch.device = DEFAULT_DEVICE,
) -> LinearProbe:
    """
    Train a logistic regression probe on the given data. This is a thin wrapped around train_multi_probe.

    Args:
        x_train: tensor of shape (num_samples, input_dim)
        y_train: tensor of shape (num_samples,), with values in [0, 1]
        batch_size: batch size for training
        num_epochs: number of epochs to train for
        lr: learning rate
        weight_decay: weight decay
        show_progress: whether to show a progress bar
        device: device to train on
    """
    return train_multi_probe(
        x_train,
        y_train.unsqueeze(1),
        num_probes=1,
        batch_size=batch_size,
        num_epochs=num_epochs,
        lr=lr,
        end_lr=end_lr,
        weight_decay=weight_decay,
        show_progress=show_progress,
        optimizer=optimizer,
        extra_loss_fn=extra_loss_fn,
        verbose=verbose,
        device=device,
    )


def _get_exponential_decay_scheduler(
    optimizer: optim.Optimizer,  # type: ignore
    start_lr: float,
    end_lr: float,
    num_steps: int,
) -> optim.lr_scheduler.ExponentialLR:
    gamma = exp(log(end_lr / start_lr) / num_steps)
    return optim.lr_scheduler.ExponentialLR(optimizer, gamma=gamma)


def _run_probe_training(
    probe: LinearProbe,
    loader: DataLoader,
    loss_fn: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    num_epochs: int,
    lr: float,
    end_lr: float,
    weight_decay: float,
    show_progress: bool,
    optimizer_name: Literal["Adam", "SGD", "AdamW"],
    extra_loss_fn: (
        Callable[[LinearProbe, torch.Tensor, torch.Tensor], torch.Tensor] | None
    ),
    verbose: bool,
    device: torch.device,
    map_acts: Callable[[torch.Tensor], torch.Tensor] | None = None,
) -> None:
    probe.train()
    if optimizer_name == "Adam":
        optimizer = optim.Adam(probe.parameters(), lr=lr, weight_decay=weight_decay)  # type: ignore
    elif optimizer_name == "SGD":
        optimizer = optim.SGD(probe.parameters(), lr=lr, weight_decay=weight_decay)  # type: ignore
    elif optimizer_name == "AdamW":
        optimizer = optim.AdamW(probe.parameters(), lr=lr, weight_decay=weight_decay)  # type: ignore
    else:
        raise ValueError(f"Unknown optimizer: {optimizer_name}")
    scheduler = _get_exponential_decay_scheduler(
        optimizer, start_lr=lr, end_lr=end_lr, num_steps=num_epochs
    )

    epoch_pbar = tqdm(range(num_epochs), disable=not show_progress, desc="Epochs")
    for epoch in epoch_pbar:
        epoch_sum_loss = 0
        batch_pbar = tqdm(
            loader,
            disable=not show_progress,
            leave=False,
            desc=f"Epoch {epoch + 1}/{num_epochs}",
        )

        for batch_embeddings, batch_labels in batch_pbar:
            if map_acts is not None:
                batch_embeddings = map_acts(batch_embeddings)
            batch_embeddings = batch_embeddings.to(device)
            batch_labels = batch_labels.to(device)
            optimizer.zero_grad()
            logits = probe(batch_embeddings)
            loss = loss_fn(logits, batch_labels)
            if extra_loss_fn is not None:
                loss += extra_loss_fn(probe, batch_embeddings, batch_labels)
            loss.backward()
            optimizer.step()

            batch_loss = loss.item()
            epoch_sum_loss += batch_loss
            batch_pbar.set_postfix({"Loss": f"{batch_loss:.8f}"})

        epoch_mean_loss = epoch_sum_loss / len(loader)
        current_lr = scheduler.get_last_lr()[0]

        epoch_pbar.set_postfix(
            {"Mean Loss": f"{epoch_mean_loss:.8f}", "LR": f"{current_lr:.2e}"}
        )

        if verbose:
            print(
                f"Epoch {epoch + 1}: Mean Loss: {epoch_mean_loss:.8f}, LR: {current_lr:.2e}"
            )

        scheduler.step()

    probe.eval()


def create_dataset_probe_training(
    vocab: list[str],
    formatter: Formatter,
    num_prompts_per_token: int,
    base_template: str,
    max_icl_examples: int = 10,
    train_test_fraction: float = 0.8,
    answer_class_fn: Callable[[str], int] = lambda answer: LETTERS.index(
        answer.strip().lower()
    ),
    # Mistral tokenizer handles the first token incorrectly if we don't do this
    prepend_separator_to_first_example: bool = True,
) -> tuple[list[tuple[SpellingPrompt, int]], list[tuple[SpellingPrompt, int]]]:
    """
    Create train and test datasets for probe training by generating prompts for each token in the given vocabulary.

    Args:
        vocab: List of tokens in the vocabulary.
        formatter: Formatter function for answers.
        num_prompts_per_token: Number of prompts to generate for each token.
        base_template: Template string for the base prompt.
        max_icl_examples: Maximum number of in-context learning examples to include.
        train_test_fraction: Fraction of vocabulary to use for training (default: 0.8).
        answer_class_fn: Function to determine the answer class from the answer string.
                         Default is to index into LETTERS for single-character answers.

    Returns:
        A tuple containing two lists of (SpellingPrompt, int) tuples for train and test sets respectively.
    """
    shuffled_vocab = random.sample(vocab, len(vocab))

    # Split into train and test vocabularies
    split_index = int(len(shuffled_vocab) * train_test_fraction)
    train_vocab = shuffled_vocab[:split_index]
    test_vocab = shuffled_vocab[split_index:]

    train_prompts, test_prompts = [], []

    def generate_prompts(token_list, examples, prompts_list):
        for token in tqdm(
            token_list,
            desc=f"Processing {'train' if prompts_list is train_prompts else 'test'} tokens",
        ):
            for _ in range(num_prompts_per_token):
                prompt = create_icl_prompt(
                    word=token,
                    examples=examples,
                    answer_formatter=formatter,
                    base_template=base_template,
                    max_icl_examples=max_icl_examples,
                    prepend_separator_to_first_example=prepend_separator_to_first_example,
                )
                answer_class = answer_class_fn(prompt.answer)
                prompts_list.append((prompt, answer_class))

    generate_prompts(train_vocab, train_vocab, train_prompts)
    generate_prompts(test_vocab, test_vocab, test_prompts)

    return train_prompts, test_prompts


def gen_and_save_df_acts_probing(
    model: HookedTransformer,
    train_dataset: list[tuple[SpellingPrompt, int]],
    test_dataset: list[tuple[SpellingPrompt, int]],
    path: str | Path,
    hook_point: str,
    batch_size: int = 64,
    position_idx: int = -2,
) -> tuple[pd.DataFrame, pd.DataFrame, np.memmap, np.memmap]:
    """
    Generate and save activations for probing tasks to the specified path

    Args:
        model: The model to use for generating activations.
        train_dataset: List of tuples containing SpellingPrompt objects and answer classes for training.
        test_dataset: List of tuples containing SpellingPrompt objects and answer classes for testing.
        path: Base path for saving outputs.
        hook_point: The model hook point to extract activations from.
        task_name: Name of the task for file naming.
        batch_size: Batch size for processing.
        position_idx: Index of the token position to extract activations from. Default is -2.

    Returns:
        A tuple containing the train and test task DataFrames and memory-mapped activation tensors.
    """
    d_model = model.cfg.d_model

    def process_dataset(dataset, prefix):
        df = pd.DataFrame(
            {
                "prompt": [prompt.base for prompt, _ in dataset],
                "HOOK_POINT": [hook_point] * len(dataset),
                "answer": [prompt.answer for prompt, _ in dataset],
                "answer_class": [answer_class for _, answer_class in dataset],
                "token": [prompt.word for prompt, _ in dataset],
            }
        )
        df.index.name = "index"

        memmap_path = os.path.join(task_dir, f"{prefix}_act_tensor.dat")
        act_tensor_memmap = np.memmap(
            memmap_path, dtype="float32", mode="w+", shape=(len(dataset), d_model)
        )

        with torch.no_grad():
            for i, batch in enumerate(
                batchify(dataset, batch_size, show_progress=True)
            ):
                batch_prompts = [prompt.base for prompt, _ in batch]
                cache = model.run_with_cache(batch_prompts, names_filter=[hook_point])[
                    1
                ]
                acts = (
                    cache[hook_point][:, position_idx, :]
                    .cpu()
                    .to(torch.float32)
                    .numpy()
                )
                start_idx = i * batch_size
                end_idx = start_idx + len(batch)
                act_tensor_memmap[start_idx:end_idx] = acts

        act_tensor_memmap.flush()
        df_path = os.path.join(task_dir, f"{prefix}_df.csv")
        df.to_csv(df_path, index=True)

        return df, act_tensor_memmap

    task_dir = path
    os.makedirs(task_dir, exist_ok=True)

    train_df, train_act_tensor = process_dataset(train_dataset, "train")
    test_df, test_act_tensor = process_dataset(test_dataset, "test")

    return train_df, test_df, train_act_tensor, test_act_tensor


def train_linear_probe_for_task(
    train_df: pd.DataFrame,
    test_df: pd.DataFrame,
    device: torch.device,
    train_activations: np.memmap,
    test_activations: np.memmap,
    num_classes: int = 26,
    batch_size: int = 4096,
    num_epochs: int = 50,
    lr: float = 1e-2,
    weight_decay=1e-4,
) -> tuple[LinearProbe, dict[str, torch.Tensor]]:
    """
    Train a linear probe for a specific task using the provided train and test DataFrames and activation tensors.
    """
    y_train = np.array(train_df["answer_class"].values)
    y_test = np.array(test_df["answer_class"].values)

    X_train_tensor = torch.from_numpy(train_activations).float().to(device)
    y_train_tensor = torch.from_numpy(y_train).long().to(device)
    X_test_tensor = torch.from_numpy(test_activations).float().to(device)
    y_test_tensor = torch.from_numpy(y_test).long().to(device)
    y_train_one_hot = one_hot(y_train_tensor, num_classes=num_classes)

    probe = train_multi_probe(
        x_train=X_train_tensor.detach().clone(),
        y_train=y_train_one_hot.detach().clone(),
        num_probes=num_classes,
        batch_size=batch_size,
        num_epochs=num_epochs,
        lr=lr,
        weight_decay=weight_decay,
        show_progress=True,
        verbose=False,
        device=device,
    )

    probe_data = {
        "X_train": X_train_tensor,
        "X_test": X_test_tensor,
        "y_train": y_train_tensor,
        "y_test": y_test_tensor,
    }

    return probe, probe_data


def save_probe_and_data(probe, probe_data, probing_path):
    task_dir = os.path.join(probing_path)
    os.makedirs(task_dir, exist_ok=True)

    probe_path = os.path.join(task_dir, "probe.pth")
    torch.save(probe, probe_path)

    data_path = os.path.join(task_dir, "data.npz")
    np.savez(
        data_path,
        X_train=probe_data["X_train"].cpu().detach().numpy(),
        X_test=probe_data["X_test"].cpu().detach().numpy(),
        y_train=probe_data["y_train"].cpu().detach().numpy(),
        y_test=probe_data["y_test"].cpu().detach().numpy(),
    )


@dataclass
class ProbeStats:
    letter: str
    f1: float
    accuracy: float
    precision: float
    recall: float


def gen_probe_stats(
    probe: LinearProbe,
    X_val: torch.Tensor,
    y_val: torch.Tensor,
    threshold: float = 0.5,
    device: torch.device = DEFAULT_DEVICE,
) -> list[ProbeStats]:
    """
    Generate statistics for a trained probe on validation data,
    treating each letter independently.

    Args:
        probe: The trained LinearProbe.
        X_val: Validation input tensor.
        y_val: Validation target tensor.
        device: The device to run computations on (default: CUDA if available, else CPU).

    Returns:
        A list of ProbeStats objects containing performance metrics for each letter.
    """

    def validator_fn(x: torch.Tensor) -> torch.Tensor:
        logits = probe(x.clone().detach().to(device))
        return (logits > threshold).float().cpu()

    results: list[ProbeStats] = []

    val_preds = validator_fn(X_val)
    y_val_cpu = y_val.cpu()

    for i, letter in enumerate(LETTERS):
        letter_preds = (val_preds[:, i]).cpu().numpy()
        letter_val_y = (y_val_cpu == i).cpu().numpy()
        results.append(
            ProbeStats(
                letter=letter,
                f1=float(f1_score(letter_val_y, letter_preds, average="binary")),
                accuracy=float(accuracy_score(letter_val_y, letter_preds)),
                precision=float(
                    precision_score(letter_val_y, letter_preds, average="binary")
                ),
                recall=float(
                    recall_score(letter_val_y, letter_preds, average="binary")
                ),
            )
        )
    return results



import random
from dataclasses import dataclass
from functools import partial
from typing import Callable

VERBOSE_FIRST_LETTER_TEMPLATE = "{word} has the first letter:"
VERBOSE_FIRST_LETTER_TOKEN_POS = -6


@dataclass
class SpellingPrompt:
    """
    Representation of a prompt used for spelling tasks. The prompt consists of a base, and answer, and a word.
    These fields might look like the following:

    base: "The word 'cat' is spelled:"
    answer: " c-a-t"
    word: "cat"

    The base may also contain ICL examples.
    """

    base: str
    answer: str
    word: str


def first_letter(
    word: str,
    prefix: str = " ",
    capitalize: bool = True,
    ignore_leading_space: bool = True,
    ignore_non_alpha_chars: bool = True,
) -> str:
    """
    return just the first letter of the word, optionally capitalized
    e.g. first_letter("cat") -> " c"
    """
    if ignore_leading_space:
        word = word.strip()
    chars = list(word)
    if ignore_non_alpha_chars:
        chars = [c for c in chars if c.isalpha()]
    first_char = chars[0]
    if capitalize:
        first_char = first_char.upper()
    return prefix + first_char


# ----- Formatters -------------------------------
Formatter = Callable[[str], str]


def first_letter_formatter(
    prefix: str = " ",
    capitalize: bool = True,
    ignore_leading_space: bool = True,
    ignore_non_alpha_chars: bool = True,
) -> Formatter:
    return partial(
        first_letter,
        prefix=prefix,
        capitalize=capitalize,
        ignore_leading_space=ignore_leading_space,
        ignore_non_alpha_chars=ignore_non_alpha_chars,
    )


# --------------------------------


def create_icl_prompt(
    word: str,
    examples: list[str],
    base_template: str = "{word}:",
    example_separator: str = "\n",
    answer_formatter: Formatter = first_letter_formatter(),
    max_icl_examples: int | None = None,
    shuffle_examples: bool = True,
    check_contamination: bool = True,
    max_attempts: int = 1000,
    prepend_separator_to_first_example: bool = False,
) -> SpellingPrompt:
    """
    Create a prompt with ICL examples in the base, optionally checking for contamination.

    Args:
        word: the word to be spelled
        examples: a list of examples to use as ICL prompts. These will be shuffled
        base_template: a string template for the base of the prompt, including "{word}" as a placeholder for the word
        example_separator: a string to use to separate the ICL examples. default is newline
        answer_formatter: a function to format the answer. default is `spelling_formatter`, which spits out a string like " c-a-t" for the word "cat"
        max_icl_examples: the maximum number of ICL examples to use. If None, all examples will be used. default is None
        shuffle_examples: whether to shuffle the examples before selecting the first `max_icl_examples`. default is True
        check_contamination: whether to check and prevent the current word from appearing in ICL examples. default is True
        max_attempts: maximum number of attempts to avoid contamination before raising an exception. default is 1000
        prepend_separator_to_first_example: whether to prepend the example separator to the first example. default is False
    """
    if max_icl_examples is None:
        max_icl_examples = len(examples)

    attempts = 0

    if check_contamination:
        while True:
            attempts += 1
            if shuffle_examples:
                icl_examples = random.sample(examples, max_icl_examples)
            else:
                icl_examples = examples[:max_icl_examples]

            if word not in icl_examples:
                break

            if attempts >= max_attempts:
                raise ValueError(
                    f"Could not find a non-contaminated set of examples after {max_attempts} attempts."
                )
    else:
        if shuffle_examples:
            icl_examples = random.sample(examples, max_icl_examples)
        else:
            icl_examples = examples[:max_icl_examples]

    icl_prompts = []
    for ex in icl_examples:
        ex_answer = answer_formatter(ex)
        ex_base = base_template.format(word=ex)
        icl_prompts.append(ex_base + ex_answer)

    word_answer = answer_formatter(word)
    word_base = base_template.format(word=word)

    base = example_separator.join(icl_prompts) + example_separator + word_base
    if prepend_separator_to_first_example:
        base = example_separator + base

    return SpellingPrompt(
        base=base,
        answer=word_answer,
        word=word,
    )


def random_icl_prompt(
    vocab: list[str],
    base_template: str = "{word}:",
    example_separator: str = "\n",
    answer_formatter: Formatter = first_letter_formatter(),
    max_icl_examples: int = 10,
    prepend_separator_to_first_example: bool = False,
) -> SpellingPrompt:
    return create_icl_prompt(
        word=random.choice(vocab),
        examples=vocab,
        base_template=base_template,
        example_separator=example_separator,
        answer_formatter=answer_formatter,
        max_icl_examples=max_icl_examples,
        prepend_separator_to_first_example=prepend_separator_to_first_example,
    )



import gc
import itertools
import json
import os
import pickle
import re
import time
from functools import partial
from itertools import permutations
from typing import Any

import numpy as np
import pandas as pd
import torch
import torch.nn.functional as F
from datasets import load_dataset
from requests.exceptions import HTTPError
from sae_lens import SAE
from tqdm import tqdm
from transformer_lens import HookedTransformer

from sae_bench.evals.unlearning.utils.intervention import (
    anthropic_clamp_resid_SAE_features,
)
from sae_bench.evals.unlearning.utils.var import (
    GEMMA_INST_FORMAT,
    MIXTRAL_INST_FORMAT,
    PRE_QUESTION_FORMAT,
    PRE_WMDP_BIO,
)

all_permutations = list(permutations([0, 1, 2, 3]))


def load_dataset_with_retries(
    dataset_path: str, dataset_name: str, split: str, retries: int = 5, delay: int = 20
):
    """
    Tries to load the dataset with a specified number of retries and delay between attempts.

    Raises:
    - HTTPError: If the dataset cannot be loaded after the given number of retries.
    """
    for attempt in range(retries):
        try:
            dataset = load_dataset(dataset_path, dataset_name, split=split)
            return dataset  # Successful load
        except HTTPError as e:
            if attempt < retries - 1:
                print(
                    f"Attempt {attempt + 1} failed: {e}. Retrying in {delay} seconds..."
                )
                time.sleep(delay)  # Wait before retrying
            else:
                print(f"Failed to load dataset after {retries} attempts.")
                raise


def calculate_MCQ_metrics(
    model: HookedTransformer,
    mcq_batch_size: int,
    artifacts_folder: str,
    dataset_name: str = "wmdp-bio",
    target_metric: str | None = None,
    question_subset: list[int] | None = None,
    question_subset_file: str | None = None,
    permutations: list[list[int]] = [[0, 1, 2, 3]],
    verbose: bool = True,
    without_question: bool = False,
    prompt_format: str | None = None,
    split: str = "all",
    **kwargs: Any,
) -> dict[str, Any]:
    """
    Calculate metrics for a multiple-choice question (MCQ) dataset using a given model.

    Parameters:
    ----------
    model : HookedTransformer
    dataset_name : str, default='wmdp-bio' - Or the dataset_name of MMLU
    target_metric : str | None - Name of the metric used to select a subset of questions
    question_subset : list[int] | None - A list of indices specifying the subset of questions to be used
    question_subset_file : str | None - Path to a file containing the indices for a subset of the questions to be used. Overrides question_subset if provided
    permutations : list[list[int]], default=[[0, 1, 2, 3]] - List of permutations to be applied to the question indices
    verbose : bool, default=True
    without_question : bool, default=False - Evaluate the model without instruction and question if True
    prompt_format : str | None - The format of the prompt to be used. Can be None, 'GEMMA_INST_FORMAT' or 'MIXTRAL_INST_FORMAT'
    split : str, default='all'
    **kwargs : Any - Additional arguments

    Returns:
    -------
    metrics : dict[str, Any] - A dictionary containing the calculated metrics for the dataset.
    """

    metrics = {}

    # Load dataset
    assert isinstance(dataset_name, str)
    if dataset_name == "wmdp-bio":
        pre_question = PRE_WMDP_BIO
        dataset = load_dataset_with_retries("cais/wmdp", "wmdp-bio", split="test")
    else:
        pre_question = PRE_QUESTION_FORMAT.format(
            subject=dataset_name.replace("_", " ")
        )
        # pre_question = 'The following are multiple choice questions (with answers) about history'
        dataset = load_dataset_with_retries("cais/mmlu", dataset_name, split="test")

    answers = [x["answer"] for x in dataset]  # type: ignore
    questions = [x["question"] for x in dataset]  # type: ignore
    choices_list = [x["choices"] for x in dataset]  # type: ignore

    # Select subset of questions
    assert target_metric in [
        None,
        "correct",
        "correct-iff-question",
        "correct_no_tricks",
        "all",
    ], "target_metric not recognised"
    assert split in ["all", "train", "test"], "split not recognised"
    if target_metric is not None:
        full_dataset_name = (
            f"mmlu-{dataset_name.replace('_', '-')}"
            if dataset_name != "wmdp-bio"
            else dataset_name
        )
        question_subset_file = (
            f"data/question_ids/{split}/{full_dataset_name}_{target_metric}.csv"
        )
        question_subset_file = os.path.join(artifacts_folder, question_subset_file)

    if question_subset_file is not None:
        question_subset = np.genfromtxt(question_subset_file, ndmin=1, dtype=int)  # type: ignore

    # Only keep desired subset of questions
    if question_subset is not None:
        answers = [answers[i] for i in question_subset if i < len(answers)]
        questions = [questions[i] for i in question_subset if i < len(questions)]
        choices_list = [
            choices_list[i] for i in question_subset if i < len(choices_list)
        ]

    # changing prompt_format
    if model.cfg.model_name in ["gemma-2-9b-it", "gemma-2-2b-it"]:
        prompt_format = "GEMMA_INST_FORMAT"
    else:
        raise Exception("Model prompt format not found.")

    if permutations is None:
        prompts = [
            convert_wmdp_data_to_prompt(
                question,
                choices,
                prompt_format=prompt_format,
                without_question=without_question,
                pre_question=pre_question,
            )
            for question, choices in zip(questions, choices_list)
        ]
    else:
        prompts = [
            [
                convert_wmdp_data_to_prompt(
                    question,
                    choices,
                    prompt_format=prompt_format,
                    permute_choices=p,
                    without_question=without_question,
                    pre_question=pre_question,
                )
                for p in permutations
            ]
            for question, choices in zip(questions, choices_list)
        ]
        prompts = [item for sublist in prompts for item in sublist]

        answers = [[p.index(answer) for p in permutations] for answer in answers]
        answers = [item for sublist in answers for item in sublist]

    actual_answers = answers

    batch_size = np.minimum(len(prompts), mcq_batch_size)
    n_batches = len(prompts) // batch_size

    if len(prompts) > batch_size * n_batches:
        n_batches = n_batches + 1

    if isinstance(model, HookedTransformer):
        output_probs = get_output_probs_abcd(
            model, prompts, batch_size=batch_size, n_batches=n_batches, verbose=verbose
        )
    else:
        output_probs = get_output_probs_abcd_hf(
            model,
            model.tokenizer,
            prompts,
            batch_size=batch_size,
            n_batches=n_batches,
            verbose=verbose,
        )

    predicted_answers = output_probs.argmax(dim=1)

    n_predicted_answers = len(predicted_answers)

    actual_answers = torch.tensor(actual_answers)[:n_predicted_answers].to(
        model.cfg.device
    )

    is_correct = (actual_answers == predicted_answers).to(torch.float)
    mean_correct = is_correct.mean()

    metrics["mean_correct"] = float(mean_correct.item())
    metrics["total_correct"] = int(np.sum(is_correct.cpu().numpy()))
    metrics["is_correct"] = is_correct.cpu().numpy()

    metrics["output_probs"] = output_probs.to(torch.float16).cpu().numpy()
    # metrics['actual_answers'] = actual_answers.cpu().numpy()

    # metrics['predicted_answers'] = predicted_answers.cpu().numpy()
    # metrics['predicted_probs'] = predicted_probs.to(torch.float16).cpu().numpy()
    # metrics['predicted_probs_of_correct_answers'] = predicted_prob_of_correct_answers.to(torch.float16).cpu().numpy()
    # metrics['mean_predicted_prob_of_correct_answers'] = float(np.mean(predicted_prob_of_correct_answers.to(torch.float16).cpu().numpy()))
    # metrics['mean_predicted_probs'] = float(np.mean(predicted_probs.to(torch.float16).cpu().numpy()))

    # unique, counts = np.unique(metrics['predicted_answers'], return_counts=True)
    # metrics['value_counts'] = dict(zip([int(x) for x in unique], [int(x) for x in counts]))

    # metrics['sum_abcd'] = metrics['output_probs'].sum(axis=1)

    return metrics


def get_output_probs_abcd(model, prompts, batch_size=2, n_batches=100, verbose=True):
    """
    Calculates probability of selecting A, B, C, & D for a given input prompt
    and language model. Returns tensor of shape (len(prompts), 4).
    """

    spaces_and_single_models = [
        "gemma-2b-it",
        "gemma-2b",
        "gemma-2-9b",
        "gemma-2-9b-it",
        "gemma-2-2b-it",
        "gemma-2-2b",
    ]
    if model.cfg.model_name in spaces_and_single_models:
        answer_strings = ["A", "B", "C", "D", " A", " B", " C", " D"]
    elif model.cfg.model_name in ["Mistral-7B-v0.1"]:
        answer_strings = ["A", "B", "C", "D"]
    else:
        raise Exception("Model name not hardcoded in this function.")

    answer_tokens = model.to_tokens(answer_strings, prepend_bos=False).flatten()

    # batch_size = 1

    with torch.no_grad():
        output_probs = []

        for i in tqdm(range(n_batches), disable=not verbose):
            prompt_batch = prompts[i * batch_size : i * batch_size + batch_size]
            current_batch_size = len(prompt_batch)

            # prepend_bos is False because the prompt already has a BOS token due to the instruct format
            token_batch = model.to_tokens(
                prompt_batch, padding_side="right", prepend_bos=False
            ).to(model.cfg.device)

            assert (token_batch == model.tokenizer.bos_token_id).sum().item() == len(
                token_batch
            )

            token_lens = [
                len(model.to_tokens(x, prepend_bos=False)[0]) for x in prompt_batch
            ]
            next_token_indices = torch.tensor([x - 1 for x in token_lens]).to(
                model.cfg.device
            )

            vals = model(token_batch, return_type="logits")
            vals = vals[
                torch.arange(current_batch_size).to(model.cfg.device),
                next_token_indices,
            ].softmax(-1)
            # vals = torch.vstack([x[i] for x, i in zip(vals, next_token_indices)]).softmax(-1)
            # vals = vals[0, -1].softmax(-1)
            vals = vals[:, answer_tokens]
            if model.cfg.model_name in spaces_and_single_models:
                vals = vals.reshape(-1, 2, 4).max(dim=1)[0]
            output_probs.append(vals)

        output_probs = torch.vstack(output_probs)

    return output_probs


def convert_wmdp_data_to_prompt(
    question,
    choices,
    prompt_format=None,
    pre_question=PRE_WMDP_BIO,
    permute_choices=None,
    without_question=False,
):
    """
    Takes in the question and choices for WMDP data and converts it to a prompt,
    including a pre-question prompt, question, answers with A, B, C & D, followed
    by "Answer:"

    datapoint: datapoint containing question and choices
    prompt_format: can be None (default), GEMMA_INST_FORMAT or MIXTRAL_INST_FORMAT
    """

    pre_answers = ["A. ", "B. ", "C. ", "D. "]
    pre_answers = ["\n" + x for x in pre_answers]
    post_answers = "\nAnswer:"

    if permute_choices is not None:
        choices = [choices[i] for i in permute_choices]

    answers = r"".join([item for pair in zip(pre_answers, choices) for item in pair])

    if prompt_format is None:
        if without_question:
            prompt = r"".join([answers, post_answers])[
                1:
            ]  # slice it to remove the '\n'
        else:
            prompt = r"".join([pre_question, question, answers, post_answers])

    elif prompt_format == "GEMMA_INST_FORMAT":
        if without_question:
            prompt = answers[1:]  # slice it to remove the '\n'
        else:
            prompt = r"".join([pre_question, question, answers])

        prompt = GEMMA_INST_FORMAT.format(prompt=prompt)
        prompt = prompt + "Answer: ("

    elif prompt_format == "MIXTRAL_INST_FORMAT":
        if without_question:
            prompt = answers[1:]  # slice it to remove the '\n'
        else:
            prompt = r"".join([pre_question, question, answers, post_answers])
        prompt = MIXTRAL_INST_FORMAT.format(prompt=prompt)
        # prompt = prompt + "Answer:"

    else:
        raise Exception("Prompt format not recognised.")

    return prompt


def get_per_token_loss(logits, tokens):
    log_probs = F.log_softmax(logits, dim=-1)
    # Use torch.gather to find the log probs of the correct tokens
    # Offsets needed because we're predicting the NEXT token (this means the final logit is meaningless)
    # None and [..., 0] needed because the tensor used in gather must have the same rank.
    predicted_log_probs = log_probs[..., :-1, :].gather(
        dim=-1, index=tokens[..., 1:, None]
    )[..., 0]
    return -predicted_log_probs


def get_output_probs_abcd_hf(
    model, tokenizer, prompts, batch_size=1, n_batches=100, verbose=True
):
    # answer_strings = ["A", "B", "C", "D"]
    answer_strings = [" A", " B", " C", " D"]
    istart = 0

    # answer_tokens = model.to_tokens(answer_strings, prepend_bos=False).flatten()
    answer_tokens = torch.tensor(
        [tokenizer(x)["input_ids"][1:] for x in answer_strings]
    ).to(model.cfg.device)

    with torch.no_grad():
        output_probs = []

        for i in tqdm(range(n_batches), disable=not verbose):
            prompt_batch = prompts[i * batch_size : i * batch_size + batch_size]
            current_batch_size = len(prompt_batch)
            token_batch = [
                torch.tensor(tokenizer(x)["input_ids"][istart:]).to(model.cfg.device)
                for x in prompt_batch
            ]
            next_token_indices = torch.tensor([len(x) - 1 for x in token_batch]).to(
                model.cfg.device
            )
            max_len = np.max([len(x) for x in token_batch])
            token_batch = [
                torch.concatenate(
                    (
                        x,
                        torch.full((max_len - len(x),), tokenizer.pad_token_id).to(
                            model.cfg.device
                        ),
                    )
                )
                for x in token_batch
            ]
            token_batch = torch.vstack(token_batch)

            logits = model(token_batch).logits
            vals = logits[torch.arange(current_batch_size), next_token_indices]
            vals = vals.softmax(-1)[:, answer_tokens]

            # if model.cfg.model_name in spaces_and_single_models:
            # vals = vals.reshape(-1, 2, 4).max(dim=1)[0]
            output_probs.append(vals)

        output_probs = torch.vstack(output_probs)
    return output_probs[:, :, 0]


def modify_model(model, sae, **ablate_params):
    model.reset_hooks()

    # Select intervention function
    if ablate_params["intervention_method"] == "scale_feature_activation":
        # ablation_method = anthropic_remove_resid_SAE_features
        raise NotImplementedError
    elif ablate_params["intervention_method"] == "remove_from_residual_stream":
        # ablation_method = remove_resid_SAE_features
        raise NotImplementedError
    elif ablate_params["intervention_method"] == "clamp_feature_activation":
        ablation_method = anthropic_clamp_resid_SAE_features
    elif ablate_params["intervention_method"] == "clamp_feature_activation_jump":
        # ablation_method = anthropic_clamp_jump_relu_resid_SAE_features
        raise NotImplementedError
    elif ablate_params["intervention_method"] == "clamp_feature_activation_random":
        # ablation_method = partial(anthropic_clamp_resid_SAE_features, random=True)
        raise NotImplementedError

    # Hook function
    features_to_ablate = ablate_params["features_to_ablate"]

    if (
        isinstance(ablate_params["features_to_ablate"], int)
        or isinstance(features_to_ablate, np.int64)  # type: ignore
        or isinstance(features_to_ablate, np.float64)  # type: ignore
    ):
        features_to_ablate = [ablate_params["features_to_ablate"]]
        ablate_params["features_to_ablate"] = features_to_ablate

    hook_params = dict(ablate_params)
    del hook_params["intervention_method"]

    ablate_hook_func = partial(ablation_method, sae=sae, **hook_params)  # type: ignore
    # features_to_ablate=features_to_ablate,
    # multiplier=ablate_params['multiplier']
    # )

    # Hook point
    if (
        "custom_hook_point" not in ablate_params
        or ablate_params["custom_hook_point"] is None
    ):
        hook_point = sae.cfg.hook_name
    else:
        hook_point = ablate_params["custom_hook_point"]

    model.add_hook(hook_point, ablate_hook_func)


def compute_loss_added(
    model, sae, activation_store, n_batch=2, split="all", verbose=False, **ablate_params
):
    """
    Computes loss added for model and SAE intervention
    """

    activation_store.iterable_dataset = iter(activation_store.dataset)

    # only take even batches for train and odd batches for test
    if split in ["train", "test"]:
        n_batch *= 2

    with torch.no_grad():
        loss_diffs = []

        for i in tqdm(range(n_batch), disable=not verbose):
            tokens = activation_store.get_batch_tokenized_data()

            # skip the irrelevant batch
            if split == "train" and i % 2 == 0:
                continue
            elif split == "test" and i % 2 == 1:
                continue

            # Compute baseline loss
            model.reset_hooks()
            baseline_loss = model(tokens, return_type="loss")

            gc.collect()
            torch.cuda.empty_cache()

            # Calculate modified loss
            model.reset_hooks()
            modify_model(model, sae, **ablate_params)
            modified_loss = model(tokens, return_type="loss")

            gc.collect()
            torch.cuda.empty_cache()

            model.reset_hooks()

            loss_diff = modified_loss.item() - baseline_loss.item()
            loss_diffs.append(loss_diff)

        return np.mean(loss_diffs)


def get_baseline_metrics(
    model: HookedTransformer,
    mcq_batch_size: int,
    artifacts_folder: str,
    dataset_name,
    metric_param,
    recompute=False,
    split="all",
):
    """
    Compute the baseline metrics or retrieve if pre-computed and saved
    """

    model.reset_hooks()

    full_dataset_name = (
        f"mmlu-{dataset_name.replace('_', '-')}"
        if dataset_name != "wmdp-bio"
        else dataset_name
    )
    q_type = metric_param["target_metric"]

    baseline_metrics_file = os.path.join(
        artifacts_folder,
        "data/baseline_metrics",
        f"{split}/{full_dataset_name}_{q_type}.json",
    )
    os.makedirs(os.path.dirname(baseline_metrics_file), exist_ok=True)

    if not recompute and os.path.exists(baseline_metrics_file):
        # Load the json
        with open(baseline_metrics_file) as f:
            baseline_metrics = json.load(f)

        # Convert lists to arrays for ease of use
        for key, value in baseline_metrics.items():
            if isinstance(value, list):
                baseline_metrics[key] = np.array(value)

        return baseline_metrics

    else:
        baseline_metrics = calculate_MCQ_metrics(
            model,
            mcq_batch_size,
            artifacts_folder,
            dataset_name=dataset_name,
            split=split,
            **metric_param,
        )

        metrics = baseline_metrics.copy()

        # Convert lists to arrays for ease of use
        for key, value in metrics.items():
            if isinstance(value, np.ndarray):
                metrics[key] = value.tolist()

        with open(baseline_metrics_file, "w") as f:
            json.dump(metrics, f)

        return baseline_metrics


def modify_and_calculate_metrics(
    model: HookedTransformer,
    mcq_batch_size: int,
    artifacts_folder: str,
    sae: SAE,
    dataset_names=["wmdp-bio"],
    metric_params={"wmdp-bio": {"target_metric": "correct"}},
    n_batch_loss_added=2,
    activation_store=None,
    split="all",
    verbose=False,
    **ablate_params,
):
    metrics_for_current_ablation = {}

    if "loss_added" in dataset_names:
        loss_added = compute_loss_added(
            model,
            sae,
            activation_store,
            n_batch=n_batch_loss_added,
            split=split,
            verbose=verbose,
            **ablate_params,
        )

        metrics_for_current_ablation["loss_added"] = loss_added
        dataset_names = [x for x in dataset_names if x != "loss_added"]

    model.reset_hooks()
    modify_model(model, sae, **ablate_params)

    for dataset_name in dataset_names:
        if dataset_name in metric_params:
            metric_param = metric_params[dataset_name]
        else:
            metric_param = {"target_metric": "correct", "verbose": verbose}

        dataset_metrics = calculate_MCQ_metrics(
            model,
            mcq_batch_size,
            artifacts_folder,
            dataset_name=dataset_name,
            split=split,
            **metric_param,
        )
        metrics_for_current_ablation[dataset_name] = dataset_metrics

    model.reset_hooks()

    return metrics_for_current_ablation


def generate_ablate_params_list(main_ablate_params, sweep):
    combinations = [
        dict(zip(sweep.keys(), values)) for values in itertools.product(*sweep.values())
    ]

    cfg_list = []
    for combo in combinations:
        specific_inputs = main_ablate_params.copy()
        specific_inputs.update(combo)
        cfg_list.append(specific_inputs)
    return cfg_list


def calculate_metrics_list(
    model: HookedTransformer,
    mcq_batch_size: int,
    sae: SAE,
    main_ablate_params,
    sweep,
    artifacts_folder: str,
    force_rerun: bool,
    dataset_names=["wmdp-bio"],
    metric_params={"wmdp-bio": {"target_metric": "correct"}},
    n_batch_loss_added=2,
    activation_store=None,
    split="all",
    target_metric="correct",
    verbose=False,
    save_metrics=False,
    save_metrics_dir=None,
    retain_threshold=None,
):
    """
    Calculate metrics for combinations of ablations
    """

    metrics_list = []

    # First get baseline metrics and ensure that target question ids exist
    baseline_metrics = {}

    for dataset_name in [x for x in dataset_names if x != "loss_added"]:
        # Ensure that target question ids exist
        save_target_question_ids(model, mcq_batch_size, artifacts_folder, dataset_name)

        if dataset_name in metric_params:
            metric_param = metric_params[dataset_name]
        else:
            metric_param = {"target_metric": target_metric, "verbose": False}

        # metrics[dataset_name] = dataset_metrics

        baseline_metric = get_baseline_metrics(
            model,
            mcq_batch_size,
            artifacts_folder,
            dataset_name,
            metric_param,
            split=split,
        )

        baseline_metrics[dataset_name] = baseline_metric

    if "loss_added" in dataset_names:
        baseline_metrics["loss_added"] = 0

    metrics_list.append(baseline_metrics)

    # Now do all ablatation combinations and get metrics each time
    ablate_params_list = generate_ablate_params_list(main_ablate_params, sweep)

    for ablate_params in tqdm(ablate_params_list):
        # check if metrics already exist
        intervention_method = ablate_params["intervention_method"]
        multiplier = ablate_params["multiplier"]
        n_features = len(ablate_params["features_to_ablate"])
        layer = sae.cfg.hook_layer

        save_file_name = f"{intervention_method}_multiplier{multiplier}_nfeatures{n_features}_layer{layer}_retainthres{retain_threshold}.pkl"
        full_path = os.path.join(save_metrics_dir, save_file_name)  # type: ignore

        if os.path.exists(full_path) and not force_rerun:
            with open(full_path, "rb") as f:
                ablated_metrics = pickle.load(f)
            metrics_list.append(ablated_metrics)
            continue

        ablated_metrics = modify_and_calculate_metrics(
            model,
            mcq_batch_size,
            artifacts_folder,
            sae,
            dataset_names=dataset_names,
            metric_params=metric_params,
            n_batch_loss_added=n_batch_loss_added,
            activation_store=activation_store,
            split=split,
            verbose=verbose,
            **ablate_params,
        )
        metrics_list.append(ablated_metrics)

        if save_metrics:
            modified_ablate_metrics = ablated_metrics.copy()
            modified_ablate_metrics["ablate_params"] = ablate_params

            os.makedirs(os.path.dirname(full_path), exist_ok=True)
            with open(full_path, "wb") as f:
                pickle.dump(modified_ablate_metrics, f)

    return metrics_list


def convert_list_of_dicts_to_dict_of_lists(list_of_dicts):
    # Initialize an empty dictionary to hold the lists
    dict_of_lists = {}

    # Iterate over each dictionary in the list
    for d in list_of_dicts:
        for key, value in d.items():
            if key not in dict_of_lists:
                dict_of_lists[key] = []
            dict_of_lists[key].append(value)

    return dict_of_lists


def create_df_from_metrics(metrics_list):
    df_data = []

    dataset_names = list(metrics_list[0].keys())

    if "loss_added" in dataset_names:
        dataset_names.remove("loss_added")

    if "ablate_params" in dataset_names:
        dataset_names.remove("ablate_params")

    for metric in metrics_list:
        if "loss_added" in metric:
            loss_added = metric["loss_added"]
        else:
            loss_added = np.NaN
        mean_correct = [
            metric[dataset_name]["mean_correct"] for dataset_name in dataset_names
        ]
        mean_predicted_probs = [
            metric[dataset_name]["mean_predicted_probs"]
            for dataset_name in dataset_names
        ]

        metric_data = np.concatenate(([loss_added], mean_correct, mean_predicted_probs))

        df_data.append(metric_data)

    df_data = np.array(df_data)

    columns = ["loss_added"] + dataset_names + [x + "_prob" for x in dataset_names]
    df = pd.DataFrame(df_data, columns=columns)  # type: ignore

    return df


def save_target_question_ids(
    model: HookedTransformer,
    mcq_batch_size: int,
    artifacts_folder: str,
    dataset_name: str,
    train_ratio: float = 0.5,
):
    """
    Find and save the question ids where the model
    1. correct: all permutations correct
    2. correct-iff-question: all permutations correct iff with instruction and questions
    3. correct-no-tricks: all permutations correct and without tricks
    """

    full_dataset_name = (
        f"mmlu-{dataset_name.replace('_', '-')}"
        if dataset_name != "wmdp-bio"
        else dataset_name
    )
    model_name = model.cfg.model_name

    # Check if the files already exist
    file_paths = [
        os.path.join(
            artifacts_folder,
            "data/question_ids",
            f"{split}/{full_dataset_name}_{q_type}.csv",
        )
        for q_type in ["correct", "correct-iff-question", "correct-no-tricks"]
        for split in ["train", "test", "all"]
    ]

    if all(os.path.exists(file_path) for file_path in file_paths):
        print(
            f"All target question ids for {model_name} on {dataset_name} already exist. No need to generate target ids."
        )
        return

    print(f"Saving target question ids for {model_name} on {dataset_name}...")

    metrics = calculate_MCQ_metrics(
        model,
        mcq_batch_size,
        artifacts_folder,
        dataset_name,
        permutations=all_permutations,  # type: ignore
    )
    metrics_wo_question = calculate_MCQ_metrics(
        model,
        mcq_batch_size,
        artifacts_folder,
        dataset_name,
        permutations=all_permutations,  # type: ignore
        without_question=True,
    )

    # find all permutations correct
    all_types = {
        "correct": (correct_ids := _find_all_permutation_correct_ans(metrics)),
        "correct-iff-question": _find_correct_iff_question(
            correct_ids, metrics_wo_question
        ),
        "correct-no-tricks": _find_correct_no_tricks(correct_ids, dataset_name),
    }

    for q_type, q_ids in all_types.items():
        train, test = _split_train_test(q_ids, train_ratio=train_ratio)
        splits = {"train": train, "test": test, "all": q_ids}

        for split, ids in splits.items():
            file_name = os.path.join(
                artifacts_folder,
                "data/question_ids",
                f"{split}/{full_dataset_name}_{q_type}.csv",
            )
            os.makedirs(os.path.dirname(file_name), exist_ok=True)
            np.savetxt(file_name, ids, fmt="%d")
            print(f"{file_name} saved, with {len(ids)} questions")


def _find_all_permutation_correct_ans(metrics):
    each_question_acc = metrics["is_correct"].reshape(-1, 24)
    questions_correct = each_question_acc.sum(axis=1) == 24
    correct_question_id = np.where(questions_correct)[0]

    return correct_question_id


def _find_correct_iff_question(correct_questions, metrics_wo_question):
    each_question_acc_wo_question = metrics_wo_question["is_correct"].reshape(-1, 24)
    correct_wo_question = np.where(each_question_acc_wo_question.sum(axis=1) == 24)[0]
    questions_correct_iff_question = list(
        set(correct_questions) - set(correct_wo_question)
    )

    return np.array(questions_correct_iff_question)


def load_dataset_from_name(dataset_name: str):
    if dataset_name == "wmdp-bio":
        dataset = load_dataset("cais/wmdp", "wmdp-bio", split="test")
    else:
        dataset = load_dataset("cais/mmlu", dataset_name, split="test")
    return dataset


def _find_correct_no_tricks(correct_questions, dataset_name):
    dataset = load_dataset_from_name(dataset_name)
    choices_list = [x["choices"] for x in dataset]  # type: ignore

    def matches_pattern(s):
        pattern = r"^(Both )?(A|B|C|D) (and|&) (A|B|C|D)$"
        return bool(re.match(pattern, s)) or s == "All of the above"

    correct_no_tricks = []
    for question_id in correct_questions:
        if not any(matches_pattern(choice) for choice in choices_list[question_id]):
            correct_no_tricks.append(question_id)

    return np.array(correct_no_tricks)


def _split_train_test(questions_ids, train_ratio=0.5):
    """shuffle then split the questions ids into train and test"""
    questions_ids = np.random.permutation(questions_ids)
    split = int(len(questions_ids) * train_ratio)
    return questions_ids[:split], questions_ids[split:]



import json

import torch
import torch.nn as nn
from huggingface_hub import hf_hub_download

import sae_bench.custom_saes.base_sae as base_sae


class TopKSAE(base_sae.BaseSAE):
    threshold: torch.Tensor
    k: torch.Tensor

    def __init__(
        self,
        d_in: int,
        d_sae: int,
        k: int,
        model_name: str,
        hook_layer: int,
        device: torch.device,
        dtype: torch.dtype,
        use_threshold: bool = False,
        hook_name: str | None = None,
    ):
        hook_name = hook_name or f"blocks.{hook_layer}.hook_resid_post"
        super().__init__(d_in, d_sae, model_name, hook_layer, device, dtype, hook_name)

        assert isinstance(k, int) and k > 0
        self.register_buffer("k", torch.tensor(k, dtype=torch.int, device=device))

        self.use_threshold = use_threshold
        if use_threshold:
            # Optional global threshold to use during inference. Must be positive.
            self.register_buffer(
                "threshold", torch.tensor(-1.0, dtype=dtype, device=device)
            )

    def encode(self, x: torch.Tensor):
        """Note: x can be either shape (B, F) or (B, L, F)"""
        post_relu_feat_acts_BF = nn.functional.relu(
            (x - self.b_dec) @ self.W_enc + self.b_enc
        )

        if self.use_threshold:
            if self.threshold < 0:
                raise ValueError(
                    "Threshold is not set. The threshold must be set to use it during inference"
                )
            encoded_acts_BF = post_relu_feat_acts_BF * (
                post_relu_feat_acts_BF > self.threshold
            )
            return encoded_acts_BF

        post_topk = post_relu_feat_acts_BF.topk(self.k, sorted=False, dim=-1)  # type: ignore

        tops_acts_BK = post_topk.values
        top_indices_BK = post_topk.indices

        buffer_BF = torch.zeros_like(post_relu_feat_acts_BF)
        encoded_acts_BF = buffer_BF.scatter_(
            dim=-1, index=top_indices_BK, src=tops_acts_BK
        )
        return encoded_acts_BF

    def decode(self, feature_acts: torch.Tensor):
        return (feature_acts @ self.W_dec) + self.b_dec

    def forward(self, x: torch.Tensor):
        x = self.encode(x)
        recon = self.decode(x)
        return recon


def load_dictionary_learning_topk_sae(
    repo_id: str,
    filename: str,
    model_name: str,
    device: torch.device,
    dtype: torch.dtype,
    layer: int | None = None,
    local_dir: str = "downloaded_saes",
    use_threshold_at_inference: bool = False,
) -> TopKSAE:
    assert "ae.pt" in filename

    path_to_params = hf_hub_download(
        repo_id=repo_id,
        filename=filename,
        force_download=False,
        local_dir=local_dir,
    )

    pt_params = torch.load(path_to_params, map_location=torch.device("cpu"))

    config_filename = filename.replace("ae.pt", "config.json")
    path_to_config = hf_hub_download(
        repo_id=repo_id,
        filename=config_filename,
        force_download=False,
        local_dir=local_dir,
    )

    with open(path_to_config) as f:
        config = json.load(f)

    if layer is not None:
        assert layer == config["trainer"]["layer"]
    else:
        layer = config["trainer"]["layer"]

    # Transformer lens often uses a shortened model name
    assert model_name in config["trainer"]["lm_name"]

    k = config["trainer"]["k"]

    # Print original keys for debugging
    print("Original keys in state_dict:", pt_params.keys())

    # Map old keys to new keys
    key_mapping = {
        "encoder.weight": "W_enc",
        "decoder.weight": "W_dec",
        "encoder.bias": "b_enc",
        "bias": "b_dec",
        "k": "k",
    }

    if "threshold" in pt_params:
        if use_threshold_at_inference:
            key_mapping["threshold"] = "threshold"
        else:
            del pt_params["threshold"]

    # Create a new dictionary with renamed keys
    renamed_params = {key_mapping.get(k, k): v for k, v in pt_params.items()}

    # due to the way torch uses nn.Linear, we need to transpose the weight matrices
    renamed_params["W_enc"] = renamed_params["W_enc"].T
    renamed_params["W_dec"] = renamed_params["W_dec"].T

    # Print renamed keys for debugging
    print("Renamed keys in state_dict:", renamed_params.keys())

    sae = TopKSAE(
        d_in=renamed_params["b_dec"].shape[0],
        d_sae=renamed_params["b_enc"].shape[0],
        k=k,
        model_name=model_name,
        hook_layer=layer,  # type: ignore
        device=device,
        dtype=dtype,
        use_threshold=use_threshold_at_inference,
    )

    sae.load_state_dict(renamed_params)  # type: ignore

    sae.to(device=device, dtype=dtype)

    d_sae, d_in = sae.W_dec.data.shape

    assert d_sae >= d_in

    if config["trainer"]["trainer_class"] == "TopKTrainer":
        sae.cfg.architecture = "topk"
    else:
        raise ValueError(f"Unknown trainer class: {config['trainer']['trainer_class']}")

    normalized = sae.check_decoder_norms()
    if not normalized:
        raise ValueError("Decoder vectors are not normalized. Please normalize them")

    return sae


if __name__ == "__main__":
    repo_id = "adamkarvonen/saebench_pythia-160m-deduped_width-2pow12_date-0104"
    filename = "TopKTrainer_EleutherAI_pythia-160m-deduped_ctx1024_0104/resid_post_layer_8/trainer_20/ae.pt"
    layer = 8

    device = "cuda" if torch.cuda.is_available() else "cpu"
    dtype = torch.float32

    model_name = "EleutherAI/pythia-160m-deduped"
    hook_name = f"blocks.{layer}.hook_resid_post"

    sae = load_dictionary_learning_topk_sae(
        repo_id,
        filename,
        model_name,
        device,  # type: ignore
        dtype,
        layer=layer,
    )
    sae.test_sae(model_name)



GEMMA_INST_FORMAT = (
    "<bos><start_of_turn>user\n{prompt}<end_of_turn>\n<start_of_turn>model\n"
)

MIXTRAL_INST_FORMAT = " [INST] {prompt} [/INST]"

# def convert_wmdp_data_to_prompt(datapoint, prompt_format=None, few_shot=False, few_shot_datapoint=None):
PRE_WMDP_BIO = (
    "The following are multiple choice questions (with answers) about biology.\n"
)
PRE_WMDP_CYBER = (
    "The following are multiple choice questions (with answers) about cyber security.\n"
)
PRE_QUESTION_FORMAT = (
    "The following are multiple choice questions (with answers) about {subject}.\n"
)
